#!/bin/bash

# Copyright (c) 2021-2023, NVIDIA CORPORATION.  All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

# This is a script to do OTA update for no partition layout change
# Usage:
#    sudo ./nv_ota_update_without_layout_change.sh

# 1. Validate target board
# 2. Enable A/B redundancy if not enabled
# 3. Check whether ROOTFS A/B is enabled or not
# 4. Update rootfs if required
#   4a. Update rootfs by running rootfs updater without booting to recovery
#       kernel if ROOTFS A/B is enabled.
#   4b. Or, update rootfs by booting to recovery kernel if ROOTFS A/B is not enabled
# 5. Update bootloader if required
set -e

CHIP_ID=
DEFAULT_ROOTFS="APP"
DEFAULT_ROOTFS_B="${DEFAULT_ROOTFS}_b"
DEVICE_NVME="nvme0n1"
EXTERNAL_DEVICE=
EXT_DEV="external_device"
INT_DEV="internal_device"
NVME_USER_DEVICE="/dev/nvme0n1"
R35_IMAGES_DIR="images-R35-ToT"
REBOOT_TO_RECOVERY=0
ROOTFS_AB_ENABLED=0
ROOTFS_IMAGE="system.img"
SDMMC_USER_DEVICE="/dev/mmcblk0"
UEFI_CAPSULE_FILE="TEGRA_BL.Cap"
UEFI_CAPSULE_FILE_AGX_ORIN_3701_000="TEGRA_BL_3701_000.Cap"
UEFI_CAPSULE_FILE_AGX_ORIN_3701_300="TEGRA_BL_3701_300.Cap"
UPDATE_BOOTLOADER=0
UPDATE_ROOTFS=0
UPDATE_SLOT=

get_update_control()
{
	# Determine whether to update bootloader only or update rootfs
	# only or update both bootloader and rootfs
	# Usage:
	#        get_update_control {work_dir} {_ret_update_bl} {_ret_update_rfs}
	local work_dir="${1}"
	local _ret_update_bl="${2}"
	local _ret_update_rfs="${3}"
	local update_control_file="${work_dir}/update_control"

	if [ ! -f "${update_control_file}" ]; then
		ota_log "The file ${update_control_file} is not found"
		return 1
	fi

	local line=
	local update_bl=0
	local update_rfs=0
	while read -r line
	do
		case ${line} in
		bootloader) update_bl=1; ;;
		rootfs) update_rfs=1; ;;
		*) break; ;;
		esac
	done <"${update_control_file}"

	if [ "${_ret_update_bl}" != "" ]; then
		eval "${_ret_update_bl}=${update_bl}"
	fi
	if [ "${_ret_update_rfs}" != "" ]; then
		eval "${_ret_update_rfs}=${update_rfs}"
	fi
	return 0
}

get_rootfs_a_b_enabled()
{
	# Check whether ROOTFS A/B is enabled
	# Usage:
	#        get_rootfs_a_b_enabled {_ret_rootfs_ab_enabled} {_ret_rootfs_current_slot}
	local _ret_rootfs_ab_enabled="${1}"
	local _ret_rootfs_current_slot="${2}"
	local rootfs_ab_enabled=
	local rootfs_current_slot=
	local _nvbootctrl=
	_nvbootctrl="$(which nvbootctrl)"

	# Use the return code to determine whether rootfs A/B is enabled
	# 0: Rootfs A/B is disabled
	# 1: Rootfs A/B is enabled and current slot is A
	# 2: Rootfs A/B is enabled and current slotis B
	local ret_code=
	"${_nvbootctrl}" -t rootfs is-rootfs-ab-enabled >/dev/null 2>&1
	ret_code=$?
	if [ "${ret_code}" == 0 ]; then
		rootfs_ab_enabled=0
		rootfs_current_slot=0
	elif [ "${ret_code}" == 1 ]; then
		rootfs_ab_enabled=1
		rootfs_current_slot=0
	elif [ "${ret_code}" == 2 ]; then
		rootfs_ab_enabled=1
		rootfs_current_slot=1
	else
		ota_log "Failed to run \"${_nvbootctrl} -t rootfs is-rootfs-ab-enabled\""
		return 1
	fi
	eval "${_ret_rootfs_ab_enabled}=${rootfs_ab_enabled}"
	eval "${_ret_rootfs_current_slot}=${rootfs_current_slot}"
	return 0
}

# Update partition with A/B redundency, such as kernel partition
update_partition_with_ab()
{
	# Update partition with A/B redundancy
	# Usage:
	#        update_partition_with_ab {work_dir} {partition}
	local work_dir="${1}"
	local partition="${2}"

	case ${CHIP_ID} in
		0x19)
			if [ "${UPDATE_SLOT}" == "B" ]; then
				partition="${partition}_b"
			fi
			;;
		0x23)
			if [ "${UPDATE_SLOT}" == "B" ]; then
				partition="B_${partition}"
			else
				partition="A_${partition}"
			fi
			;;
		*)
			ota_log "Unsupported chip id: ${CHIP_ID}"
			return 1
			;;
	esac

	# Update partition
	local target_dir=
	if [ -n "${EXTERNAL_DEVICE}" ]; then
		target_dir="${work_dir}/${EXT_DEV}/${R35_IMAGES_DIR}"
	else
		target_dir="${work_dir}/${INT_DEV}/${R35_IMAGES_DIR}"
	fi
	ota_log "install_partition_with_ab ${target_dir} ${partition} ${OTA_DEVICE}"
	if ! install_partition_with_ab "${target_dir}" "${partition}" "${OTA_DEVICE}"; then
		ota_log "Failed to run \"install_partition_with_ab ${target_dir} ${partition} ${OTA_DEVICE}\""
		return 1
	fi
	return 0
}

# Update partition with alt partition, such as recovery partition
update_partition_with_alt()
{
	# Update partition with alt partition
	# Usage:
	#        update_partition_with_alt {work_dir} {partition}
	local work_dir="${1}"
	local partition="${2}"

	# Update partition
	local target_dir=
	if [ -n "${EXTERNAL_DEVICE}" ]; then
		target_dir="${work_dir}/${EXT_DEV}/${R35_IMAGES_DIR}"
	else
		target_dir="${work_dir}/${INT_DEV}/${R35_IMAGES_DIR}"
	fi

	ota_log "install_partition_with_alt ${target_dir} ${partition}"
	if ! install_partition_with_alt "${target_dir}" "${partition}"; then
		ota_log "Failed to run \"install_partition_with_alt ${target_dir} ${partition}\""
		return 1
	fi
	return 0
}

restore_extlinux_conf()
{
	# Restore extlinux.conf
	# Usage:
	#        restore_extlinux_conf {root_partition} {base_version}
	local root_partition="${1}"
	local base_version="${2}"
	local tmp_mnt=/tmp/root_tmp
	local extlinux_conf="${tmp_mnt}/boot/extlinux/extlinux.conf"
	local extlinux_conf_backup="${extlinux_conf}.${base_version}"

	mkdir -p "${tmp_mnt}"
	if ! mount "${root_partition}" "${tmp_mnt}"; then
		ota_log "Failed to mount ${root_partition} on ${tmp_mnt}"
		return 1
	fi
	if [ ! -f "${extlinux_conf_backup}" ]; then
		ota_log "The extlinux.conf backup(${extlinux_conf_backup}) is not found"
		return 1
	fi
	cp -f "${extlinux_conf_backup}" "${extlinux_conf}"

	# For OTA from R35.2.1 on t194 devices, add the "video=efifb:off"
	# to the end of the APPEND line in the restored exlinux.conf
	if [ "${base_version}" == "R35-2" ] && [ "${CHIP_ID}" == "0x19" ]; then
		ota_log "Adding video=efifb:off to the end of the APPEND line"
		sed -i "/^[ \t]*APPEND/s|\$| video=efifb:off|" "${extlinux_conf}"
	fi

	if ! umount "${tmp_mnt}"; then
		ota_log "Failed to umount ${tmp_mnt}"
		return 1
	fi
	rm -rf "${tmp_mnt}"
	sync
	return 0
}

update_rootfs_runtime()
{
	# Update rootfs on non-current runtime without rebooting
	# to recovery kernel
	# Usage:
	#        update_rootfs_runtime {work_dir} {update_rootfs_slot} {dev}
	local work_dir="${1}"
	local update_rootfs_slot="${2}"
	local dev="${3}"
	if [ "${ROOTFS_AB_ENABLED}" != 1 ]; then
		return 1
	fi

	local rootfs_image="${work_dir}/${dev}/${ROOTFS_IMAGE}"
	if [ ! -f "${rootfs_image}" ]; then
		ota_log "The rootfs image ${rootfs_image} is not found"
		return 1
	fi

	# Set rootfs updater
	local rootfs_updater=
	if ! set_rootfs_updater "${work_dir}" rootfs_updater; then
		ota_log "Failed to run \"set_rootfs_updater ${work_dir} rootfs_updater\""
		return 1
	fi

	# Set rootfs name according to the "${update_rootfs_slot}"
	local rootfs_name=
	local rootfs_part_devnode=
	local rootfs_storage=
	if [ "${update_rootfs_slot}" == "B" ]; then
		rootfs_name="${DEFAULT_ROOTFS_B}"
	else
		rootfs_name="${DEFAULT_ROOTFS}"
	fi
	if [ "${dev}" == "${EXT_DEV}" ]; then
		rootfs_storage="${NVME_USER_DEVICE}"
	else
		rootfs_storage="${SDMMC_USER_DEVICE}"
	fi
	get_devnode_from_name "${rootfs_name}" "${rootfs_storage}" rootfs_part_devnode
	if [ ! -e "${rootfs_part_devnode}" ];  then
		ota_log "Failed to find the ${rootfs_name} partition"
		return 1
	fi

	# Run rootfs updater to update rootfs
	local cmd="${rootfs_updater} -p ${rootfs_part_devnode} -d ${work_dir} ${rootfs_image}"
	ota_log "Updating rootfs by running command: ${cmd}"
	if ! eval "${cmd}"; then
		ota_log "Failed to run command \"${cmd}\""
		return 1
	fi

	# Restore extlinux.conf
	local base_version=
	base_version="$(cat "${work_dir}/base_version")"
	ota_log "restore_extlinux_conf ${rootfs_part_devnode} ${base_version}"
	if ! restore_extlinux_conf "${rootfs_part_devnode}" "${base_version}"; then
		ota_log "Failed to restore extlinux.conf on ${rootfs_part_devnode}" "${base_version}"
		return 1
	fi

	return 0
}

switch_current_rootfs()
{
	# Switch to boot from the non-current rootfs that is updated
	# Usage:
	#        switch_current_rootfs
	local _nvbootctrl=
	local update_slot=
	local ret_code=

	if [ "${UPDATE_SLOT}" == "B" ]; then
		update_slot=1
	else
		update_slot=0
	fi
	_nvbootctrl="$(which nvbootctrl)"
	"${_nvbootctrl}" set-active-boot-slot "${update_slot}" >/dev/null 2>&1
	ret_code=$?
	if [ "${ret_code}" == 0 ]; then
		ota_log "Switch to updated non-current rootfs(${UPDATE_SLOT})"
		return 0
	else
		ota_log "Failed to switch to updated non-current rootfs(${UPDATE_SLOT})"
		return 1
	fi
}

update_bootloader()
{
	# Update bootloader partitions
	# Usage:
	#        update_bootloader {work_dir}
	local work_dir="${1}"

	# Update bootloader by triggering UEFI capsule update
	# If need to reboot to recovery kernel, then will trigger UEFI
	# capsule update in recovery kernel
	if [ "${REBOOT_TO_RECOVERY}" == 0 ]; then
		ota_log "trigger_uefi_capsule_update ${OTA_WORK_DIR} ${TARGET_BOARD}" "${OTA_DEVICE}"
		if ! trigger_uefi_capsule_update "${OTA_WORK_DIR}" "${TARGET_BOARD}" "${OTA_DEVICE}"; then
			ota_log "Failed to run \"trigger_uefi_capsule_update ${OTA_WORK_DIR} ${TARGET_BOARD} ${OTA_DEVICE}\""
			exit 1
		fi
	fi
	ota_log "Bootloader on non-current slot(${UPDATE_SLOT}) is to be updated once device is rebooted"
}

update_rootfs_with_a_b_enabled()
{
	# Update rootfs partition when ROOTFS A/B is enabled.
	# Also, update kernel and kernel-dtb partitions on the
	# non-current chain.
	# Usage:
	#        update_rootfs_with_a_b_enabled {work_dir}
	local work_dir="${1}"
	local update_rootfs_slot=
	local dev=

	# Update rootfs on non-current slot by running rootfs updater
	if [ -n "${EXTERNAL_DEVICE}" ]; then
		dev="${EXT_DEV}"
	else
		dev="${INT_DEV}"
	fi
	ota_log "update_rootfs_runtime ${work_dir} ${UPDATE_SLOT} ${dev}"
	if ! update_rootfs_runtime "${work_dir}" "${UPDATE_SLOT}" "${dev}"; then
		ota_log "Failed to run \"update_rootfs_runtime ${work_dir} ${UPDATE_SLOT} ${dev}\""
		return 1
	fi
	ota_log "Rootfs on non-current slot(${UPDATE_SLOT}) is updated"

	# Update kernel and kerne-dtb partitions
	ota_log "update_partition_with_ab ${work_dir} kernel"
	if ! update_partition_with_ab "${work_dir}" "kernel"; then
		ota_log "Failed to run \"update_partition_with_ab ${work_dir} kernel\""
		return 1
	fi
	ota_log "update_partition_with_ab ${work_dir} kernel-dtb"
	if ! update_partition_with_ab "${work_dir}" "kernel-dtb"; then
		ota_log "Failed to run \"update_partition_with_ab ${work_dir} kernel-dtb\""
		return 1
	fi

	# Switch current rootfs if only rootfs is updated
	if [ "${UPDATE_BOOTLOADER}" == 0 ]; then
		ota_log "switch_current_rootfs"
		if ! switch_current_rootfs; then
			ota_log "Failed to run \"switch_current_rootfs\""
			exit 1
		fi
		ota_log "Will switch to boot from rootfs(${UPDATE_SLOT}) once device is rebooted"
	fi

	# Clean rootfs chain status
	ota_log "Rest chain status for rootfs(${UPDATE_SLOT})"
	if ! reset_rootfs_chain_status "${UPDATE_SLOT}" "${TARGET_BOARD}"; then
		ota_log "Failed to run \"reset_rootfs_chain_status ${UPDATE_SLOT} ${TARGET_BOARD}\""
		exit 1
	fi
}

update_rootfs_with_a_b_disabled()
{
	# Update rootfs partition when ROOTFS A/B is disabled
	# Usage:
	#        update_rootfs_with_a_b_disabled {work_dir}
	local work_dir="${1}"

	# Force booting to recovery kernel on next boot by writing UEFI variable
	ota_log "force_booting_to_recovery ${TARGET_BOARD}"
	if ! force_booting_to_recovery "${TARGET_BOARD}"; then
		ota_log "Failed to run \"force_booting_to_recovery ${TARGET_BOARD}\""
		return 1
	fi
	REBOOT_TO_RECOVERY=1
	ota_log "Rootfs is to be updated in recovery kernel once device is rebooted."
}

update_rootfs()
{
	# Update rootfs partition
	# Usage:
	#        update_rootfs {work_dir}
	local work_dir="${1}"

	# If rootfs A/B is not enabled, boot to recovery kernel and then
	# update rootfs in recovery kernel.
	# If rootfs A/B is enabled, update rootfs on the non-current slot
	# without booting to recovery kernel.
	if [ "${ROOTFS_AB_ENABLED}" != 1 ]; then
		ota_log "update_rootfs_with_a_b_disabled ${work_dir}"
		if ! update_rootfs_with_a_b_disabled "${work_dir}"; then
			ota_log "Failed to run \"update_rootfs_with_a_b_disabled ${work_dir}\""
			return 1
		fi
	else
		ota_log "update_rootfs_with_a_b_enabled ${work_dir}"
		if ! update_rootfs_with_a_b_enabled "${work_dir}"; then
			ota_log "Failed to run \"update_rootfs_with_a_b_enabled ${work_dir}\""
			return 1
		fi
	fi
}

update_misc_partitions()
{
	# Update misc partitions including kernel, kernel-dtb,
	# recovery, recovery-dtb and esp
	# Usage:
	#        update_misc_partitions {work_dir}
	local work_dir="${1}"

	# Update recovery, recovery-dtb
	ota_log "update_partition_with_alt ${work_dir} recovery"
	if ! update_partition_with_alt "${work_dir}" "recovery"; then
		ota_log "Failed to run \"update_partition_with_alt ${work_dir} recovery\""
		return 1
	fi
	ota_log "update_partition_with_alt ${work_dir} recovery-dtb"
	if ! update_partition_with_alt "${work_dir}" "recovery-dtb"; then
		ota_log "Failed to run \"update_partition_with_alt ${work_dir} recovery-dtb\""
		return 1
	fi

	# Update esp
	ota_log "update_partition_with_alt ${work_dir} esp"
	if ! update_partition_with_alt "${work_dir}" "esp"; then
		ota_log "Failed to run \"update_partition_with_alt ${work_dir} esp\""
		return 1
	fi
}

select_uefi_capsule()
{
	# Select the correct UEFI capsule according to the boardrev
	# Usage:
	#        select_uefi_capsule {work_dir}
	local work_dir="${1}"
	local base_version=
	base_version="$(cat "${work_dir}"/base_version)"

	# If base version is not R35.2 or board is not jetson-agx-orin-devkit,
	# return directly as the valid "TEGRA_BL.Cap" is under "${work_dir}"
	if [ "${TARGET_BOARD}" != "jetson-agx-orin-devkit" ] \
		|| [ "${base_version}" != "R35-2" ] ; then
		return 0
	fi

	local nv_boot_control_conf="/etc/nv_boot_control.conf"
	local boardver=
	boardver="$(grep -m 1 TNSPEC < "${nv_boot_control_conf}" | cut -d\  -f 2 | cut -d- -f 2)"
	if [ "${boardver}" == "" ]; then
		ota_log "Failed to get boardver from TNSPEC in ${nv_boot_control_conf}"
		return 1
	fi

	local uefi_capsule=
	if [[ "${boardver}" == "000" ]] || [[ "${boardver}" == TS* ]] \
		|| [[ "${boardver}" == EB* ]] \
		|| [[ ($((boardver)) -gt 0 && $((boardver)) -lt 300) ]]; then
		ota_log "Using ${UEFI_CAPSULE_FILE_AGX_ORIN_3701_000} as UEFI capsule"
		uefi_capsule="${work_dir}/${UEFI_CAPSULE_FILE_AGX_ORIN_3701_000}"
	else
		ota_log "Using ${UEFI_CAPSULE_FILE_AGX_ORIN_3701_300} as UEFI capsule"
		uefi_capsule="${work_dir}/${UEFI_CAPSULE_FILE_AGX_ORIN_3701_300}"
	fi
	cp "${uefi_capsule}" "${work_dir}/${UEFI_CAPSULE_FILE}"
	return 0
}

# Make sure that this script is running in root privilege
USERID=$(id -u)
if [ "${USERID}" -ne 0 ]; then
       ota_log "Please run this program as root."
       exit 0
fi

if [ $# -ne 1 ]; then
	usage
fi

nargs=$#;
OTA_DEVICE=${!nargs};

ota_log "Command: ${BASH_SOURCE[0]}"

if [ ! -e "${OTA_DEVICE}" ]; then
	echo "Invalid disk path ${OTA_DEVICE}"
	usage
fi

source ./nv_ota_check_version.sh
source ./nv_ota_common.func
source ./nv_ota_customer.conf
source ./ota_multi_board_specs.sh
source ./nv_ota_update_alt_part.func

# Check whether current board matches with OTA package target board
ota_log "check_target_board ${OTA_WORK_DIR} TARGET_BOARD"
if ! check_target_board "${OTA_WORK_DIR}" "TARGET_BOARD"; then
	ota_log "Failed to run \"check_target_board ${OTA_WORK_DIR} TARGET_BOARD\""
	exit 1
fi

# Get the chip id
ota_log "get_chip_id CHIP_ID"
if ! get_chip_id "CHIP_ID"; then
	ota_log "Failed to run \"get_chip_id CHIP_ID\""
	exit 1
fi

# Support NVMe device as external device
if [ "${OTA_DEVICE}" == "${NVME_USER_DEVICE}" ]; then
	EXTERNAL_DEVICE="${DEVICE_NVME}"
fi

# The board specific images are stored under images-XXX-XXX/<board_spec>/ directory.
# Need to parse current target board spec and then move the images from
# corresponding board_spec directory to image-XXX-XXX.
ota_log "ota_choose_images ${OTA_WORK_DIR}"
if ! ota_choose_images "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"ota_choose_images ${OTA_WORK_DIR}\""
	exit 1
fi

# Check whether ROOTFS A/B is enabled
ota_log "get_rootfs_a_b_enabled ROOTFS_AB_ENABLED ROOTFS_CURRENT_SLOT"
if ! get_rootfs_a_b_enabled "ROOTFS_AB_ENABLED" "ROOTFS_CURRENT_SLOT"; then
	ota_log "Failed to run \"get_rootfs_a_b_enabled ROOTFS_AB_ENABLED ROOTFS_CURRENT_SLOT\""
	exit 1
fi
ota_log "ROOTFS_AB_ENABLED=${ROOTFS_AB_ENABLED}"
ota_log "ROOTFS_CURRENT_SLOT=${ROOTFS_CURRENT_SLOT}"

# Get the slot to updated
ota_log "get_update_slot UPDATE_SLOT"
if ! get_update_slot "UPDATE_SLOT"; then
	ota_log "Failed to run \"get_update_slot UPDATE_SLOT\""
	exit 1
fi
ota_log "UPDATE_SLOT=${UPDATE_SLOT}"

# Get the value of "update_control" file to determine whether
# to update bootloader or/and rootfs
ota_log "get_update_control ${OTA_WORK_DIR} UPDATE_BOOTLOADER UPDATE_ROOTFS"
if ! get_update_control "${OTA_WORK_DIR}" "UPDATE_BOOTLOADER" "UPDATE_ROOTFS"; then
	ota_log "Failed to run \"get_update_control ${OTA_WORK_DIR} UPDATE_BOOTLOADER UPDATE_ROOTFS\""
	exit 1
fi
ota_log "UPDATE_BOOTLOADER=${UPDATE_BOOTLOADER}, UPDATE_ROOTFS=${UPDATE_ROOTFS}"

# Select UEFI capsule
ota_log "select_uefi_capsule ${OTA_WORK_DIR}"
if ! select_uefi_capsule "${OTA_WORK_DIR}"; then
	ota_log "Failed to run \"select_uefi_capsule ${OTA_WORK_DIR}\""
	exit 1
fi

# Update partitions and rootfs on storage device
if [ "${UPDATE_ROOTFS}" == 1 ]; then
	# Make sure the base version built in OTA package matches with version in rootfs.
	ota_log "check_bsp_version ${OTA_WORK_DIR} BASE_VERSION"
	if ! check_bsp_version "${OTA_WORK_DIR}" "BASE_VERSION"; then
		ota_log "Failed to run \"check_bsp_version ${OTA_WORK_DIR} BASE_VERSION\""
		exit 1
	fi

	ota_log "update_misc_partitions ${OTA_WORK_DIR}"
	if ! update_misc_partitions "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"update_misc_partitions ${OTA_WORK_DIR}\""
		exit 1
	fi

	ota_log "update_rootfs ${OTA_WORK_DIR}"
	if ! update_rootfs "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"update_rootfs ${OTA_WORK_DIR}\""
		exit 1
	fi
fi

# Update bootloader on the non-current slot if UPDATE_BOOTLOADER is 1
if [ "${UPDATE_BOOTLOADER}" == 1 ]; then
	ota_log "check_bootloader_version ${OTA_WORK_DIR}"
	if ! check_bootloader_version "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"check_bootloader_version ${OTA_WORK_DIR}\""
		exit 1
	fi

	ota_log "update_bootloader ${OTA_WORK_DIR}"
	if ! update_bootloader "${OTA_WORK_DIR}"; then
		ota_log "Failed to run \"update_bootloader ${OTA_WORK_DIR}\""
		exit 1
	fi
fi

